#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
##
#         http://www.apache.org/licenses/LICENSE-2.0
##
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import logging

from django.conf import settings
from openstack_dashboard.api import base
from watcherclient import client as wc
from watcherclient.common import api_versioning
from watcherclient import exceptions as wc_exc

LOG = logging.getLogger(__name__)

# Service type used by Watcher in the catalog
WATCHER_SERVICE = 'infra-optim'

# Default/minimal microversion
MIN_DEFAULT = '1.0'

# Microversion enabling audit start/end time fields
MV_START_END = '1.1'

# Microversion enabling skip action support
MV_SKIP_ACTION = '1.5'


def get_client(request, required=MIN_DEFAULT):
    """Return a watcher client pinned to the given microversion.

    :param request: The current Django HTTP request.
    :param required: Microversion string (e.g. ``'1.0'``,
        ``'1.1'``).
    """
    endpoint = base.url_for(request, WATCHER_SERVICE)
    insecure = settings.OPENSTACK_SSL_NO_VERIFY
    ca_file = settings.OPENSTACK_SSL_CACERT

    return wc.get_client(
        '1',
        watcher_url=endpoint,
        insecure=insecure,
        ca_file=ca_file,
        username=request.user.username,
        os_auth_token=request.user.token.id,
        os_infra_optim_api_version=required,
    )


def get_max_version(request):
    """Discover the server's maximum supported microversion.

    Queries the Watcher API root endpoint and extracts the
    ``max_version`` field from the response body.

    :param request: The current Django HTTP request.
    :returns: The max version string, or ``None`` on failure.
    """
    try:
        client = get_client(request)
        _resp, body = client.http_client.json_request(
            'GET', '/')
        version_info = (
            body.get('versions') or
            body.get('version') or
            body.get('default_version') or
            {})
        if isinstance(version_info, list) and version_info:
            version_info = version_info[0]
        if isinstance(version_info, dict):
            return version_info.get('max_version')
    except wc_exc.ClientException:
        LOG.debug('Microversion discovery failed',
                  exc_info=True)
    return None


def is_microversion_supported(max_ver, required):
    """Check whether the server supports the required microversion.

    :param max_ver: The server's max version string from
        :func:`get_max_version`, or ``None``.
    :param required: The microversion string to check
        (e.g. ``'1.1'``).
    :returns: ``True`` if the server's max version is at least
        *required*, ``False`` otherwise.
    """
    if max_ver is None:
        return False
    try:
        return (api_versioning.APIVersion(max_ver) >=
                api_versioning.APIVersion(required))
    except (wc_exc.UnsupportedVersion, TypeError):
        return False



# Copyright (c) 2016 b<>com
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from openstack_dashboard.test.test_data import exceptions
from watcherclient.common.apiclient import exceptions as wexceptions


def data(TEST):
    TEST.exceptions = exceptions.data

    watcher_exception = wexceptions.ClientException
    TEST.exceptions.watcher = exceptions.create_stubbed_exception(
        watcher_exception)



# Copyright (c) 2016 b<>com
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json
import logging

from django.http import JsonResponse
from django.urls import reverse
from django.urls import reverse_lazy
from django.utils.safestring import mark_safe
from django.utils.translation import gettext_lazy as _
import horizon.exceptions
from horizon import forms
import horizon.tables
import horizon.tabs
from horizon.utils import memoized
import horizon.workflows
import yaml

from watcher_dashboard.api import watcher
from watcher_dashboard.common import client as common_client
from watcher_dashboard.content.action_plans import tables as action_plan_tables
from watcher_dashboard.content.audits import forms as wforms
from watcher_dashboard.content.audits import tables
from watcher_dashboard.content.audits import tabs as wtabs

LOG = logging.getLogger(__name__)


class IndexView(horizon.tables.DataTableView):
    table_class = tables.AuditsTable
    template_name = 'infra_optim/audits/index.html'
    page_title = _("Audits")

    @memoized.memoized_method
    def max_version(self):
        return common_client.get_max_version(self.request)

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        create_action = {
            'name': _("New Audit"),
            'url': reverse('horizon:admin:audits:create'),
            'icon': 'fa-plus',
            'ajax_modal': True,
        }
        context['header_actions'] = [create_action]
        context['audits_count'] = self.get_audits_count()
        return context

    def get_data(self):
        audits = []
        search_opts = self.get_filters()
        try:
            server_version = self.max_version()
            version = (
                common_client.MV_START_END
                if common_client.is_microversion_supported(
                    server_version, common_client.MV_START_END)
                else None)
            audits = watcher.Audit.list(
                self.request,
                api_version=version,
                **search_opts)
        except Exception:
            horizon.exceptions.handle(
                self.request,
                _("Unable to retrieve audit information."))
        return audits

    def get_audits_count(self):
        return len(self.get_data())

    def get_filters(self):
        filters = {}
        filter_action = self.table._meta._filter_action
        if filter_action:
            filter_field = self.table.get_filter_field()
            if filter_action.is_api_filter(filter_field):
                filter_string = self.table.get_filter_string()
                if filter_field and filter_string:
                    filters[filter_field] = filter_string
        return filters


class CreateView(forms.ModalFormView):
    form_class = wforms.CreateForm
    form_id = "create_audit_form"
    modal_header = _("Create Audit")
    template_name = 'infra_optim/audits/create.html'
    success_url = reverse_lazy("horizon:admin:audits:index")
    page_title = _("Create Audit")
    submit_label = _("Create Audit")
    submit_url = reverse_lazy("horizon:admin:audits:create")


class DetailView(horizon.tables.MultiTableView):
    table_classes = (action_plan_tables.RelatedActionPlansTable,)
    tab_group_class = wtabs.AuditDetailTabs
    template_name = 'infra_optim/audits/details.html'
    redirect_url = 'horizon:admin:audits:index'
    page_title = _("Audit Details: {{ audit.uuid }}")

    @memoized.memoized_method
    def max_version(self):
        return common_client.get_max_version(self.request)

    @memoized.memoized_method
    def _get_data(self):
        audit_uuid = None
        try:
            audit_uuid = self.kwargs['audit_uuid']
            server_version = self.max_version()
            version = (
                common_client.MV_START_END
                if common_client.is_microversion_supported(
                    server_version, common_client.MV_START_END)
                else None)
            audit = watcher.Audit.get(
                self.request, audit_uuid,
                api_version=version)
        except Exception:
            msg = _('Unable to retrieve details for audit "%s".') \
                % audit_uuid
            horizon.exceptions.handle(
                self.request, msg,
                redirect=self.redirect_url)
        return audit

    def _render_pretty_parameters(self, params):
        """Return a human-friendly rendering of parameters.

        It is used to render parameters on the audit details page
        in a YAML format for readability.

        Rules:
        - If params is a JSON string, parse it first.
        - If result is a dict/list, pretty-print as YAML in a <pre> block.
        - If result is a scalar (str/int/bool), return it as-is.
        - On any error, fall back to the original params value.
        """
        try:
            obj = params
            # Parameters may be stored as a JSON-serialized string. Try to
            # decode it but tolerate non-JSON strings (e.g. plain text).
            if isinstance(params, str):
                try:
                    obj = json.loads(params)
                except Exception:
                    obj = params

            # Dicts/lists are presented as indented YAML for readability.
            if isinstance(obj, dict | list):
                dumped = yaml.safe_dump(
                    obj,
                    default_flow_style=False,
                    sort_keys=False,
                )
                return mark_safe(  # noqa: S308  # nosec B703,B308
                    f'<pre style="margin:0">{dumped}</pre>'
                )

            # Scalars or unknown types: return directly.
            if obj is not None:
                return obj
        except Exception:
            # Any unexpected issue: show the raw parameters.
            return params

    def get_related_action_plans_data(self):
        try:
            action_plan = self._get_data()
            audits = watcher.ActionPlan.list(self.request,
                                             audit=action_plan.uuid)
        except Exception as exc:
            LOG.exception(exc)
            audits = []
            msg = _('Action plan list cannot be retrieved.')
            horizon.exceptions.handle(self.request, msg)
        return audits

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        audit = self._get_data()
        context["audit"] = audit
        # Prepare pretty parameters rendering (YAML) for the template. The
        # helper encapsulates the logic so it is easier to test/maintain.
        context["audit_parameters_pretty"] = self._render_pretty_parameters(
            getattr(audit, 'parameters', None)
        )
        return context

    def get_tabs(self, request, *args, **kwargs):
        audit = self._get_data()
        # ports = self._get_ports()
        return self.tab_group_class(request, audit=audit, **kwargs)


def get_strategy_parameters(request):
    """AJAX endpoint to get strategy parameters based on audit template."""
    try:
        audit_template_uuid = request.GET.get('audit_template_uuid')
        if not audit_template_uuid:
            return JsonResponse(
                {'error': 'Audit template UUID is required'},
                status=400)

        # Get the audit template
        audit_template = watcher.AuditTemplate.get(
            request, audit_template_uuid)

        if not audit_template.strategy_uuid:
            return JsonResponse({
                'strategy_name': audit_template.strategy_name or 'auto',
                'parameters_spec': {},
                'message': ('No specific strategy selected. Parameters will '
                            'be automatically determined.')
            })

        # Get the strategy details
        strategy = watcher.Strategy.get(request, audit_template.strategy_uuid)

        # Parse parameters_spec if it exists
        parameters_spec = {}
        if hasattr(strategy, 'parameters_spec') and strategy.parameters_spec:
            if isinstance(strategy.parameters_spec, dict):
                properties = strategy.parameters_spec.get('properties', {})
                parameters_spec = properties
            elif isinstance(strategy.parameters_spec, str):
                try:
                    parsed_spec = json.loads(strategy.parameters_spec)
                    parameters_spec = parsed_spec.get('properties', {})
                except ValueError:
                    parameters_spec = {}

        return JsonResponse({
            'strategy_name': watcher.get_strategy_display_name(strategy),
            'strategy_uuid': strategy.uuid,
            'parameters_spec': parameters_spec
        })

    except Exception as e:
        LOG.exception("Error getting strategy parameters")
        return JsonResponse({'error': str(e)}, status=500)



# Copyright 2025,  Red Hat, Inc.
# All rights reserved.

# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from unittest import mock

from watcher_dashboard import api
from watcher_dashboard.common import client as common_client
from watcher_dashboard.test import helpers as test


class WatcherClientAPITests(test.APITestCase):

    def test_audit_create_defaults_to_min_version(self):
        with mock.patch('watcher_dashboard.api.watcher.watcherclient') as wc:
            client_mock = mock.Mock()
            wc.return_value = client_mock
            client_mock.audit.create = mock.Mock(return_value={})

            api.watcher.Audit.create(self.request,
                                     audit_template_uuid='tpl',
                                     audit_type='continuous',
                                     name='n',
                                     interval='60')

            wc.assert_called_with(self.request, api_version=None)
            client_mock.audit.create.assert_called_once()

    def test_audit_create_passes_explicit_api_version(self):
        with mock.patch('watcher_dashboard.api.watcher.watcherclient') as wc:
            client_mock = mock.Mock()
            wc.return_value = client_mock
            client_mock.audit.create = mock.Mock(return_value={})

            api.watcher.Audit.create(self.request,
                                     audit_template_uuid='tpl',
                                     audit_type='continuous',
                                     name='n',
                                     interval='60',
                                     start_time='2025-01-01T10:00:00',
                                     end_time='2025-01-01T11:00:00',
                                     api_version='1.1')

            wc.assert_called_with(self.request, api_version='1.1')
            client_mock.audit.create.assert_called_once()

    def test_audit_list_passes_explicit_api_version(self):
        with mock.patch('watcher_dashboard.api.watcher.watcherclient') as wc:
            client_mock = mock.Mock()
            wc.return_value = client_mock
            client_mock.audit.list = mock.Mock(return_value=[])

            api.watcher.Audit.list(self.request, api_version='1.1')

            wc.assert_called_with(self.request, api_version='1.1')

    def test_audit_get_passes_explicit_api_version(self):
        with mock.patch('watcher_dashboard.api.watcher.watcherclient') as wc:
            client_mock = mock.Mock()
            wc.return_value = client_mock
            client_mock.audit.get = mock.Mock(return_value={})

            api.watcher.Audit.get(self.request, 'audit-uuid',
                                  api_version='1.1')

            wc.assert_called_with(self.request, api_version='1.1')

    def test_is_microversion_supported(self):
        self.assertTrue(
            common_client.is_microversion_supported('1.1', '1.0'))
        self.assertTrue(
            common_client.is_microversion_supported('1.1', '1.1'))
        self.assertFalse(
            common_client.is_microversion_supported('1.0', '1.1'))
        self.assertFalse(
            common_client.is_microversion_supported(None, '1.1'))

    def test_is_microversion_supported_start_end(self):
        self.assertTrue(
            common_client.is_microversion_supported(
                '1.1', common_client.MV_START_END))
        self.assertTrue(
            common_client.is_microversion_supported(
                '1.5', common_client.MV_START_END))
        self.assertFalse(
            common_client.is_microversion_supported(
                '1.0', common_client.MV_START_END))
        self.assertFalse(
            common_client.is_microversion_supported(
                None, common_client.MV_START_END))

    def test_is_microversion_supported_skip_action(self):
        self.assertTrue(
            common_client.is_microversion_supported(
                '1.5', common_client.MV_SKIP_ACTION))
        self.assertTrue(
            common_client.is_microversion_supported(
                '2.0', common_client.MV_SKIP_ACTION))
        self.assertFalse(
            common_client.is_microversion_supported(
                '1.4', common_client.MV_SKIP_ACTION))
        self.assertFalse(
            common_client.is_microversion_supported(
                '1.1', common_client.MV_SKIP_ACTION))
        self.assertFalse(
            common_client.is_microversion_supported(
                None, common_client.MV_SKIP_ACTION))



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import logging

from django.utils.translation import gettext_lazy as _
from horizon import forms
from horizon import messages

from watcher_dashboard.api import watcher
from watcher_dashboard.common import client as common_client
from watcher_dashboard.common import exceptions as watcher_exc

LOG = logging.getLogger(__name__)


class SkipActionForm(forms.SelfHandlingForm):
    """Form to skip an action or update its skip reason."""

    action_id = forms.CharField(widget=forms.HiddenInput())
    reason = forms.CharField(
        label=_("Reason for skipping"),
        required=False,
        max_length=200,
        widget=forms.Textarea(attrs={'rows': 3}),
        help_text=_("Optional reason for skipping this action."))

    def clean(self):
        """Validate the action exists and is in a skippable state."""
        cleaned = super().clean()
        action_id = cleaned.get('action_id')
        if not action_id:
            # Django calls clean() even when required fields fail
            # validation. Without this guard, None would be passed
            # to the API as the action ID.
            return cleaned
        action = watcher.Action.get(
            self.request, action_id)
        if action is None:
            raise forms.ValidationError(
                _('Unable to retrieve action.'))
        if action.state not in ('PENDING', 'SKIPPED'):
            raise forms.ValidationError(
                _('Action must be in PENDING or SKIPPED '
                  'state.'))
        reason = cleaned.get('reason', '').strip()
        if action.state == 'SKIPPED' and not reason:
            raise forms.ValidationError(
                _('A reason is required when updating a '
                  'skipped action.'))
        cleaned['_action'] = action
        return cleaned

    def handle(self, request, data):
        """Submit skip request to the Watcher API."""
        action_id = data['action_id']
        reason = data.get('reason', '').strip() or None
        action = data['_action']

        state = None if action.state == 'SKIPPED' else 'SKIPPED'
        try:
            result = watcher.Action.update(
                request, action_id, state=state, reason=reason,
                api_version=common_client.MV_SKIP_ACTION)
        except watcher_exc.WatcherDashboardException as exc:
            LOG.info("Skip action validation error: %s", exc)
            messages.error(request, str(exc))
            return False

        if result is None:
            messages.warning(
                request,
                _('Skip is not supported by this server.'))
            return False

        if state is None:
            msg = _('Skip reason was successfully updated.')
        else:
            msg = _('Action was successfully skipped.')
        messages.success(request, msg)
        return True



# Copyright (c) 2016 b<>com
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from unittest import mock

from watcher_dashboard import api
from watcher_dashboard.test import helpers as test


class WatcherAPITests(test.APITestCase):

    def test_goal_list(self):
        goals = {'goals': self.api_goals.list()}
        watcherclient = self.stub_watcherclient()
        watcherclient.goal.list = mock.Mock(
            return_value=goals)

        ret_val = api.watcher.Goal.list(self.request)
        self.assertIsInstance(ret_val, dict)
        self.assertIn('goals', ret_val)
        for n in ret_val['goals']:
            self.assertIsInstance(n, dict)
        watcherclient.goal.list.assert_called_with(
            detail=True)

    def test_goal_get(self):
        goal = self.api_goals.first()
        goal_id = self.api_goals.first()['uuid']

        watcherclient = self.stub_watcherclient()
        watcherclient.goal.get = mock.Mock(
            return_value=goal)

        ret_val = api.watcher.Goal.get(self.request, goal_id)
        self.assertIsInstance(ret_val, dict)
        watcherclient.goal.get.assert_called_with(
            goal_id)

    def test_strategy_list(self):
        strategies = {'strategies': self.api_strategies.list()}
        watcherclient = self.stub_watcherclient()

        watcherclient.strategy.list = mock.Mock(
            return_value=strategies)

        ret_val = api.watcher.Strategy.list(self.request)
        self.assertIn('strategies', ret_val)
        for n in ret_val['strategies']:
            self.assertIsInstance(n, dict)
        watcherclient.strategy.list.assert_called_with(
            detail=True)

    def test_strategy_get(self):
        strategy = self.api_strategies.first()
        strategy_id = self.api_strategies.first()['uuid']

        watcherclient = self.stub_watcherclient()
        watcherclient.strategy.get = mock.Mock(
            return_value=strategy)

        ret_val = api.watcher.Strategy.get(self.request, strategy_id)
        self.assertIsInstance(ret_val, dict)
        watcherclient.strategy.get.assert_called_with(
            strategy_id)

    def test_audit_template_list(self):
        audit_templates = {
            'audit_templates': self.api_audit_templates.list()}
        watcherclient = self.stub_watcherclient()

        watcherclient.audit_template.list = mock.Mock(
            return_value=audit_templates)

        ret_val = api.watcher.AuditTemplate.list(self.request)

        self.assertIn('audit_templates', ret_val)
        for n in ret_val['audit_templates']:
            self.assertIsInstance(n, dict)
        watcherclient.audit_template.list.assert_called_with(
            detail=True)

    def test_audit_template_list_with_filters(self):
        search_opts = {'name': 'Audit Template 1'}
        audit_templates = {
            'audit_templates': self.api_audit_templates.filter(**search_opts)}
        watcherclient = self.stub_watcherclient()

        watcherclient.audit_template.list = mock.Mock(
            return_value=audit_templates)

        ret_val = api.watcher.AuditTemplate.list(
            self.request, **search_opts)

        self.assertIn('audit_templates', ret_val)
        for n in ret_val['audit_templates']:
            self.assertIsInstance(n, dict)

        self.assertEqual(ret_val, audit_templates)
        watcherclient.audit_template.list.assert_called_with(
            detail=True, **search_opts)

    def test_audit_template_get(self):
        audit_template = self.api_audit_templates.first()
        audit_template_id = self.api_audit_templates.first()['uuid']

        watcherclient = self.stub_watcherclient()
        watcherclient.audit_template.get = mock.Mock(
            return_value=audit_template)

        ret_val = api.watcher.AuditTemplate.get(self.request,
                                                audit_template_id)
        self.assertIsInstance(ret_val, dict)
        watcherclient.audit_template.get.assert_called_with(
            audit_template_id=audit_template_id)

    def test_audit_template_create(self):
        audit_template = self.api_audit_templates.first()
        name = audit_template['name']
        goal = audit_template['goal_uuid']
        strategy = audit_template['strategy_uuid']
        description = audit_template['description']
        scope = audit_template['scope']

        watcherclient = self.stub_watcherclient()
        watcherclient.audit_template.create = mock.Mock(
            return_value=audit_template)

        ret_val = api.watcher.AuditTemplate.create(
            self.request, name, goal, strategy,
            description, scope)
        self.assertIsInstance(ret_val, dict)
        watcherclient.audit_template.create.assert_called_with(
            name=name,
            goal=goal,
            strategy=strategy,
            description=description,
            scope=scope)

    def test_audit_template_patch(self):
        audit_template = self.api_audit_templates.first()
        audit_template_id = self.api_audit_templates.first()['uuid']
        form_data = {'name': 'new Audit Template 1'}

        watcherclient = self.stub_watcherclient()
        watcherclient.audit_template.patch = mock.Mock(
            return_value=audit_template)

        ret_val = api.watcher.AuditTemplate.patch(
            self.request, audit_template_id,
            form_data)
        self.assertIsInstance(ret_val, dict)
        watcherclient.audit_template.patch.assert_called_with(
            audit_template_id,
            [{'name': 'name', 'value': 'new Audit Template 1'}]
        )

    def test_audit_template_delete(self):
        audit_template_list = self.api_audit_templates.list()
        audit_template_id = self.api_audit_templates.first()['uuid']
        deleted_at_list = self.api_audit_templates.delete()

        watcherclient = self.stub_watcherclient()
        watcherclient.audit_template.delete = mock.Mock()
        api.watcher.AuditTemplate.delete(self.request,
                                         audit_template_id)
        self.assertEqual(audit_template_list, deleted_at_list)
        self.assertEqual(len(audit_template_list), len(deleted_at_list))
        watcherclient.audit_template.delete.assert_called_with(
            audit_template_id=audit_template_id)

    def test_audit_list(self):
        audits = {'audits': self.api_audits.list()}

        watcherclient = self.stub_watcherclient()

        watcherclient.audit.list = mock.Mock(
            return_value=audits)

        ret_val = api.watcher.Audit.list(self.request)

        self.assertIn('audits', ret_val)
        for n in ret_val['audits']:
            self.assertIsInstance(n, dict)
        watcherclient.audit.list.assert_called_with(
            detail=True)

    def test_audit_get(self):
        audit = self.api_audits.first()
        audit_id = self.api_audits.first()['uuid']

        watcherclient = self.stub_watcherclient()
        watcherclient.audit.get = mock.Mock(
            return_value=audit)

        ret_val = api.watcher.Audit.get(self.request, audit_id)
        self.assertIsInstance(ret_val, dict)
        watcherclient.audit.get.assert_called_with(
            audit=audit_id)

    def test_audit_create(self):
        audit = self.api_audits.first()
        audit_template_id = self.api_audit_templates.first()['uuid']

        audit_type = self.api_audits.first()['audit_type']
        audit_name = self.api_audits.first()['name']
        audit_template_uuid = audit_template_id

        watcherclient = self.stub_watcherclient()
        watcherclient.audit.create = mock.Mock(
            return_value=audit)

        ret_val = api.watcher.Audit.create(
            self.request, audit_template_uuid, audit_type, audit_name)
        self.assertIsInstance(ret_val, dict)
        watcherclient.audit.create.assert_called_with(
            audit_template_uuid=audit_template_uuid,
            audit_type=audit_type, auto_trigger=False, name=audit_name)

    def test_audit_create_with_interval(self):
        audit = self.api_audits.list()[1]
        audit_template_id = self.api_audit_templates.first()['uuid']

        audit_type = self.api_audits.first()['audit_type']
        audit_name = self.api_audits.first()['name']
        interval = audit['interval']
        audit_template_uuid = audit_template_id

        watcherclient = self.stub_watcherclient()
        watcherclient.audit.create = mock.Mock(
            return_value=audit)

        ret_val = api.watcher.Audit.create(
            self.request, audit_template_uuid, audit_type, audit_name,
            False, interval)
        self.assertIsInstance(ret_val, dict)
        watcherclient.audit.create.assert_called_with(
            audit_template_uuid=audit_template_uuid,
            audit_type=audit_type,
            auto_trigger=False,
            interval=interval,
            name=audit_name)

    def test_audit_create_with_auto_trigger(self):
        audit = self.api_audits.list()[1]
        audit_template_id = self.api_audit_templates.first()['uuid']

        audit_type = self.api_audits.first()['audit_type']
        audit_name = self.api_audits.first()['name']
        audit_template_uuid = audit_template_id

        watcherclient = self.stub_watcherclient()
        watcherclient.audit.create = mock.Mock(
            return_value=audit)

        ret_val = api.watcher.Audit.create(
            self.request, audit_template_uuid, audit_type, audit_name, True)
        self.assertIsInstance(ret_val, dict)
        watcherclient.audit.create.assert_called_with(
            audit_template_uuid=audit_template_uuid,
            audit_type=audit_type,
            auto_trigger=True,
            name=audit_name)

    def test_audit_create_with_parameters(self):
        audit = self.api_audits.first()
        audit_template_id = self.api_audit_templates.first()['uuid']

        audit_type = self.api_audits.first()['audit_type']
        audit_name = self.api_audits.first()['name']
        audit_template_uuid = audit_template_id
        parameters = {'memory_threshold': 0.8, 'cpu_threshold': 0.9}

        watcherclient = self.stub_watcherclient()
        watcherclient.audit.create = mock.Mock(
            return_value=audit)

        ret_val = api.watcher.Audit.create(
            self.request, audit_template_uuid, audit_type, audit_name,
            False, None, parameters)
        self.assertIsInstance(ret_val, dict)
        watcherclient.audit.create.assert_called_with(
            audit_template_uuid=audit_template_uuid,
            audit_type=audit_type,
            auto_trigger=False,
            name=audit_name,
            parameters=parameters)

    def test_audit_create_with_start_end_time(self):
        audit = self.api_audits.list()[1]
        audit_template_id = self.api_audit_templates.first()['uuid']

        audit_type = self.api_audits.first()['audit_type']
        audit_name = self.api_audits.first()['name']
        interval = audit['interval']
        start_time = '2025-01-01 10:00:00'
        end_time = '2025-01-02 18:30:00'
        audit_template_uuid = audit_template_id

        watcherclient = self.stub_watcherclient()
        watcherclient.audit.create = mock.Mock(
            return_value=audit)

        ret_val = api.watcher.Audit.create(
            self.request, audit_template_uuid, audit_type, audit_name,
            False, interval, start_time=start_time, end_time=end_time)
        self.assertIsInstance(ret_val, dict)
        watcherclient.audit.create.assert_called_with(
            audit_template_uuid=audit_template_uuid,
            audit_type=audit_type,
            auto_trigger=False,
            interval=interval,
            name=audit_name,
            start_time=start_time,
            end_time=end_time)

    def test_audit_create_with_complex_parameters(self):
        audit = self.api_audits.first()
        audit_template_id = self.api_audit_templates.first()['uuid']

        audit_type = self.api_audits.first()['audit_type']
        audit_name = self.api_audits.first()['name']
        audit_template_uuid = audit_template_id
        # Test complex JSON parameters like arrays and objects
        parameters = {
            'memory_threshold': 0.8,
            'enable_migration': True,
            'compute_nodes': [{'src_node': 's01', 'dst_node': 'd01'}],
            'excluded_instances': ['instance1', 'instance2']
        }

        watcherclient = self.stub_watcherclient()
        watcherclient.audit.create = mock.Mock(
            return_value=audit)

        ret_val = api.watcher.Audit.create(
            self.request, audit_template_uuid, audit_type, audit_name,
            False, None, parameters)
        self.assertIsInstance(ret_val, dict)
        watcherclient.audit.create.assert_called_with(
            audit_template_uuid=audit_template_uuid,
            audit_type=audit_type,
            auto_trigger=False,
            name=audit_name,
            parameters=parameters)

    def test_audit_show_with_parameters(self):
        """Test that audit detail view shows parameters when present"""
        audit = self.api_audits.first()
        audit_id = audit['uuid']
        # Add parameters to the audit data
        audit_with_params = dict(audit)
        audit_with_params['parameters'] = {
            'memory_threshold': 0.8,
            'cpu_threshold': 0.9
        }

        watcherclient = self.stub_watcherclient()
        watcherclient.audit.get = mock.Mock(
            return_value=audit_with_params)

        ret_val = api.watcher.Audit.get(self.request, audit_id)
        self.assertIsInstance(ret_val, dict)
        expected_params = {'memory_threshold': 0.8, 'cpu_threshold': 0.9}
        self.assertEqual(ret_val['parameters'], expected_params)
        watcherclient.audit.get.assert_called_with(
            audit=audit_id)

    def test_audit_delete(self):
        audit_id = self.api_audits.first()['uuid']

        watcherclient = self.stub_watcherclient()
        watcherclient.audit.delete = mock.Mock()

        api.watcher.Audit.delete(self.request, audit_id)
        watcherclient.audit.delete.assert_called_with(
            audit=audit_id)

    def test_audit_cancel(self):
        audit_id = self.api_audits.first()['uuid']

        watcherclient = self.stub_watcherclient()
        watcherclient.audit.update = mock.Mock()

        api.watcher.Audit.cancel(self.request, audit_id)
        watcherclient.audit.update.assert_called_with(
            audit=audit_id, patch=[{'op': 'replace', 'path': '/state',
                                    'value': 'CANCELLED'}])

    def test_action_plan_list(self):
        action_plans = {'action_plans': self.api_action_plans.list()}

        watcherclient = self.stub_watcherclient()

        watcherclient.action_plan.list = mock.Mock(
            return_value=action_plans)

        ret_val = api.watcher.ActionPlan.list(self.request)

        self.assertIn('action_plans', ret_val)
        for n in ret_val['action_plans']:
            self.assertIsInstance(n, dict)
        watcherclient.action_plan.list.assert_called_with(
            detail=True)

    def test_action_plan_get(self):
        action_plan = self.api_action_plans.first()
        action_plan_id = self.api_action_plans.first()['uuid']

        watcherclient = self.stub_watcherclient()
        watcherclient.action_plan.get = mock.Mock(
            return_value=action_plan)

        ret_val = api.watcher.ActionPlan.get(self.request, action_plan_id)
        self.assertIsInstance(ret_val, dict)
        watcherclient.action_plan.get.assert_called_with(
            action_plan_id=action_plan_id)

    def test_action_plan_start(self):
        action_plan_id = self.api_action_plans.first()['uuid']

        watcherclient = self.stub_watcherclient()
        watcherclient.action_plan.start = mock.Mock()

        api.watcher.ActionPlan.start(self.request, action_plan_id)
        watcherclient.action_plan.start.assert_called_with(
            action_plan_id)

    def test_action_plan_delete(self):
        action_plan_id = self.api_action_plans.first()['uuid']

        watcherclient = self.stub_watcherclient()
        watcherclient.action_plan.delete = mock.Mock()

        api.watcher.ActionPlan.delete(self.request, action_plan_id)
        watcherclient.action_plan.delete.assert_called_with(
            action_plan_id=action_plan_id)

    def test_action_list(self):
        actions = {'actions': self.api_actions.list()}
        watcherclient = self.stub_watcherclient()

        watcherclient.action.list = mock.Mock(
            return_value=actions)

        ret_val = api.watcher.Action.list(self.request)

        self.assertIn('actions', ret_val)
        for n in ret_val['actions']:
            self.assertIsInstance(n, dict)
        watcherclient.action.list.assert_called_with(
            detail=True)

    def test_action_get_with_api_version(self):
        action = self.api_actions.first()
        action_id = action['uuid']

        watcherclient = self.stub_watcherclient()
        watcherclient.action.get = mock.Mock(
            return_value=action)

        with mock.patch(
                'watcher_dashboard.api.watcher.watcherclient',
                autospec=True) as wc:
            wc.return_value = watcherclient
            api.watcher.Action.get(
                self.request, action_id, api_version='1.5')
            wc.assert_called_with(
                self.request, api_version='1.5')

    def test_action_update_state_only(self):
        action_id = self.api_actions.first()['uuid']
        with mock.patch(
                'watcher_dashboard.api.watcher.watcherclient',
                autospec=True) as wc:
            client_mock = mock.Mock()
            wc.return_value = client_mock
            client_mock.action.update = mock.Mock(
                return_value={})
            api.watcher.Action.update(
                self.request, action_id,
                state='SKIPPED', api_version='1.5')
            client_mock.action.update.assert_called_with(
                action_id=action_id,
                patch=[{'op': 'replace', 'path': '/state',
                        'value': 'SKIPPED'}])

    def test_action_update_state_and_reason(self):
        action_id = self.api_actions.first()['uuid']
        with mock.patch(
                'watcher_dashboard.api.watcher.watcherclient',
                autospec=True) as wc:
            client_mock = mock.Mock()
            wc.return_value = client_mock
            client_mock.action.update = mock.Mock(
                return_value={})
            api.watcher.Action.update(
                self.request, action_id,
                state='SKIPPED', reason='Not needed',
                api_version='1.5')
            client_mock.action.update.assert_called_with(
                action_id=action_id,
                patch=[
                    {'op': 'replace', 'path': '/state',
                     'value': 'SKIPPED'},
                    {'op': 'replace', 'path': '/status_message',
                     'value': 'Not needed'},
                ])

    def test_action_update_no_args_raises(self):
        from watcher_dashboard.common import exceptions as watcher_exc
        action_id = self.api_actions.first()['uuid']
        self.assertRaises(
            watcher_exc.WatcherDashboardException,
            api.watcher.Action.update,
            self.request, action_id)

    def test_action_update_method_not_allowed(self):
        from watcherclient.common.apiclient import (
            exceptions as wc_exc)
        action_id = self.api_actions.first()['uuid']
        with mock.patch(
                'watcher_dashboard.api.watcher.watcherclient',
                autospec=True) as wc:
            client_mock = mock.Mock()
            wc.return_value = client_mock
            client_mock.action.update = mock.Mock(
                side_effect=wc_exc.MethodNotAllowed())
            result = api.watcher.Action.update(
                self.request, action_id,
                state='SKIPPED', api_version='1.5')
            self.assertIsNone(result)

    def test_get_strategy_display_name(self):
        strategy = api.watcher.Strategy({
            'uuid': 'test-uuid',
            'name': 'zone_migration',
            'display_name': 'Zone Migration Strategy',
        })
        self.assertEqual(
            api.watcher.get_strategy_display_name(strategy),
            'Zone Migration Strategy')

    def test_get_strategy_display_name_fallback(self):
        strategy = api.watcher.Strategy({
            'uuid': 'test-uuid',
            'name': 'zone_migration',
            'display_name': '',
        })
        self.assertEqual(
            api.watcher.get_strategy_display_name(strategy),
            'zone_migration')

    def test_get_strategy_display_name_none_fallback(self):
        strategy = api.watcher.Strategy({
            'uuid': 'test-uuid',
            'name': 'zone_migration',
            'display_name': None,
        })
        self.assertEqual(
            api.watcher.get_strategy_display_name(strategy),
            'zone_migration')

    def test_get_strategy_display_name_missing_attrs(self):
        strategy = api.watcher.Strategy({'uuid': 'test-uuid'})
        self.assertRaises(
            AttributeError,
            api.watcher.get_strategy_display_name,
            strategy)
